/* The Computer Language Benchmarks Game
   http://shootout.alioth.debian.org/

   contributed by Greg Buchholz (original iterative version)
   modified by Yannick Gingras (made tail recursive)
   adjusted slightly for Utah CS 4960-01 by Matthew Flatt
*/

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

struct color { int r, g, b; } cols[]
= { { 255, 0, 0 },
    { 0, 255, 0 },
    { 0, 0, 255 },
    { 255, 255, 0 },
    { 255, 0, 255 },
    { 0, 255, 255 } };

static int const root = 0;

int iterate(double Zr, double Zi,
	    double Tr, double Ti,
	    double Cr, double Ci,
	    int iter)
{

  if (iter == 0)
    return 1;

  Zi = 2.0 * Zr * Zi + Ci;
  Zr = Tr - Ti + Cr;
  Tr = Zr * Zr;
  Ti = Zi * Zi;

  if (Tr + Ti > 4.0)
    return 0;

  return iterate (Zr, Zi, Tr, Ti, Cr, Ci, iter - 1);
}

int main (int argc, char **argv)
{
  int me, sz;
  int w, h, buflen, rowlen;
  char *buf;
  int y;
  double wrat, hrat;

  MPI_Init(&argc, &argv);

  MPI_Comm_size(MPI_COMM_WORLD, &sz);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);

  w = h = atoi (argv[1]);
  wrat = 2.0 / w;
  hrat = 2.0 / h;
  rowlen = w * 3;
  buflen = rowlen * h;
  buf = (char *) malloc (buflen);

  /* The mandelbort image is written to buf, which is an array
     of rowlen * h bytes, where each rowlen bytes is one line
     of the image (representing w pixels). */

  if (me == root)
    printf ("P6\n%d %d\n255\n", w, h);

  for (y = me; y < h; y += sz)
    {
      double Ci = y * hrat - 1.0;
      int x;
      char *pbuf = buf + (rowlen * y);

      for (x = 0; x < w; ++x)
        {
          double Cr = x * wrat - 1.5;
            
          if (iterate (0.0, 0.0, 0.0, 0.0, Cr, Ci, 50)) {
            *(pbuf++) = cols[me].r;
            *(pbuf++) = cols[me].g;
            *(pbuf++) = cols[me].b;
          } else {
            *(pbuf++) = cols[me].r / 2;
            *(pbuf++) = cols[me].g / 2;
            *(pbuf++) = cols[me].b / 2;
          }
        }

      if (me != root)
        MPI_Bsend(buf + (rowlen * y), rowlen, MPI_CHAR, root, 0, MPI_COMM_WORLD);
    }

  if (me == root) {
    int i;
    for (i = 1; i < sz; i++) {
      for (y = i; y < h; y += sz) {
        char *pbuf = buf + (rowlen * y);
        MPI_Status s;

        MPI_Recv(pbuf, rowlen, MPI_CHAR, i, 0, MPI_COMM_WORLD, &s);
      }
    }

    fwrite (buf, buflen, 1, stdout);
  }

  return 0;
}
